#include	<stdio.h>
#include	<math.h>
#include	"gene.h"
#define	MODE2	0

double	drand48();
extern	double	CellSize;
double	MoveNut = 0.02, IncNut = 0.1, ThGlucose = 0.2, ThH2O = 0.2;
double	DryAway = 0.01;
#define	H2O0	0.1
#define	H2Oa	0.01

cell_inc_nut(c) Cell *c;
{
#if MODE2
	if (c->y > 0) {
		double	d, d2;
		c->h2o *= (1. - DryAway);
		d = (1. - c->glu) * IncNut;
		d2 = c->h2o * IncNut;
		if (d > d2) d = d2;
		c->glu += d; c->h2o -= d;
	} else c->h2o += (1. - c->h2o)
		* ((1.-H2O0)*(1.-exp(H2Oa*c->y/CellSize))+H2O0) * IncNut;
#else
	if (c->y > 0) {
		c->glu += (1. - c->glu) * IncNut;
		c->h2o *= (1. - DryAway);
	} else c->h2o += (1. - c->h2o) * IncNut;
#endif
}
cell_move_nut(c) Cell *c;
{
	c->newglu = c->glu; c->newh2o = c->h2o;
	if (c->mother) {
		double	d;
		d = (c->glu - c->mother->glu) * MoveNut;
		c->newglu -= d; c->mother->newglu += d;
		d = (c->h2o - c->mother->h2o) * MoveNut;
		c->newh2o -= d; c->mother->newh2o += d;
	}
}
cell_step(ind,c)
Individual	*ind;
Cell	*c;
{
	Gene	*action;
	double	ax[4][4];
	c->glu = c->newglu; c->h2o = c->newh2o;
#ifndef NOMETABOLISM
	if (c->glu < ThGlucose || c->h2o < ThH2O) return 0;
#endif
	c->rr = 1.;
	if (c->state & 0xf8) return 0;
	action = ind->gene + (c->state & 0x07);
	daughter_pose(c,action,ax);
	if (new_cell(ind,ax,(action->state>>4)&0x0f,c)) {
		c->state = action->state & 0x0f;
		c->glu -= ThGlucose; c->h2o -= ThH2O;
	} else	c->state |= 0x10;
	return 1;
}

daughter_pose(m,gene,a)
Cell	*m;		/* mother cell */
Gene	*gene;		/* spawning direction */
double	a[4][4];	/* daughter's matrix */
{
	int	i, j, k;
	double	r, d[3][3];
	double	x, ca, sa, cb, sb, cc, sc;
	bzero(a,sizeof(double)*16);
	x = (double)gene->alpha * M_PI / 127.5;
	ca = cos(x); sa = sin(x);
	x = (double)gene->beta * M_PI / 127.5;
	cb = cos(x); sb = sin(x);
	x = (double)gene->gamma * M_PI / 127.5;
	cc = cos(x); sc = sin(x);
	d[0][0] = cb * cc;
	d[0][1] = cb * sc;
	d[0][2] = 0;
	d[1][0] = sa * sb * cc - ca * sc;
	d[1][1] = sa * sb * sc + ca * cc;
	d[1][2] = sa * cb;
	d[2][0] = ca * sb * cc + sa * sc;
	d[2][1] = ca * sb * sc - sa * cc;
	d[2][2] = ca * cb;
	for (i = 0; i < 3; i ++) for (j = 0; j < 3; j ++)
		for (k = 0; k < 3; k ++) a[i][j] += m->afm[i][k] * d[k][j];
	for (i = 0; i < 3; i ++) a[3][i] = m->afm[3][i];
	for (i = 0; i < 4; i ++)
		a[i][3] = m->afm[i][3] + CellSize * 2 * a[i][1];
}

new_cell(ind,ax,st,mother)
Individual	*ind;
double	ax[4][4];
int	st;
Cell	*mother;
{
	Cell	*c, *alloc_cell();
	double	x = ax[0][3], y = ax[1][3], z = ax[2][3];
	extern	unsigned long	Step;
	extern	int	MaxCells;
	if (ind->size >= MaxCells) return 1;
	if (mother) {
		double r = CellSize * 2 /
		sqrt(pow(x-mother->x,2.0)+pow(y-mother->y,2.0)
		+pow(z-mother->z,2.0));
		x = mother->x + (x - mother->x) * r;
		y = mother->y + (y - mother->y) * r;
		z = mother->z + (z - mother->z) * r;
	}
	if (if_conflict(ind,x,y,z)) return 0;
	c = alloc_cell(ind);
	if (mother) c->glu = c->h2o = c->rr = 0.;
	else c->glu = c->h2o = c->rr = 1.;
	c->x = x; c->y = y; c->z = z;
	memcpy(c->afm,ax,sizeof(double)*16);
	c->state = st;
	c->step = Step;
	c->mother = mother;
	ind->size ++;
	return 1;
}
