#include	<stdio.h>
#include	<math.h>
#include	<Xm/RowColumn.h>
#include	<Xm/CascadeB.h>
#include	<Xm/DrawingA.h>
#include	<Xm/Text.h>
#include	"gene.h"
#define	NNODE	9
#define	ARROWHEAD	0.6
#define	STRINGFONT	"*-lucida-bold-r-normal-sans-34-*"
#define	HIDDENCOLOR	"grey80"

extern	Individual	*Population, *MonitoredInd;
extern	Widget	app_sh, GenomeGraphButton;
extern	Display	*dpy;
extern	Arg	ShellArgs[];
extern	int	NShellArg;
static	Widget	shell = NULL, message;
static	Dimension	WinWidth,WinHeight;
static	Pixel	fg, bg, hidden_pixel;
static	Window	window;
static	GC	xgc;
static	XFontStruct	*fst = NULL;
static	int	dia, BaseOffset;

typedef struct { int x,y,next,daughter,active,col; double th; } Node;
static	Node	node[NNODE];
typedef enum { ArcNext, ArcDaughter } ArcType;

static void draw_string(x,y,str)
int	x, y;
char	*str;
{
	int	w;
	w = XTextWidth(fst,str,strlen(str)) / 2;
	XDrawString(dpy,window,xgc,x-w,y+BaseOffset,str,strlen(str));
}
static void draw_arrow(from,to,arc)
int	from, to;
ArcType	arc;
{
	int	x1,y1,x2,y2,dx,dy;
	double	d;
	XPoint	points[3];
	switch (arc) {
		case ArcNext:
		XSetLineAttributes(dpy,xgc,3,LineSolid,CapNotLast,JoinMiter);
		break;
		case ArcDaughter:
		XSetLineAttributes(dpy,xgc,5,LineOnOffDash,CapNotLast,JoinMiter);
	}
	x1 = node[from].x; y1 = node[from].y;
	if (from == to) {
		x1 += (cos(node[to].th)*1.4142 - 1.) * dia/2;
		y1 += (sin(node[to].th)*1.4142 - 1.) * dia/2;
		XDrawArc(dpy,window,xgc,x1,y1,dia,dia,
			(int)((1.25-node[to].th/M_PI)*180*64),270*64);
	} else {
		x2 = node[to].x; y2 = node[to].y;
		d = dia * .5 / hypot((double)(dx=x1-x2),(double)(dy=y1-y2));
		x1 -= dx * d; x2 += dx * d;
		y1 -= dy * d; y2 += dy * d; 
		XDrawLine(dpy,window,xgc,x1,y1,x2,y2);
		points[0].x = x2;
		points[0].y = y2;
		points[1].x = dx * d * ARROWHEAD - dy * d * ARROWHEAD/3.;
		points[1].y = dy * d * ARROWHEAD + dx * d * ARROWHEAD/3.;
		points[2].x = dy * d * ARROWHEAD/1.5;
		points[2].y = - dx * d * ARROWHEAD/1.5;
		XFillPolygon(dpy,window,xgc,points,3,Convex,CoordModePrevious);
	}
}
static void draw_node(i)
int	i;
{
	int	k;
	char	ns[4];
	XSetLineAttributes(dpy,xgc,3,LineSolid,CapNotLast,JoinMiter);
	XDrawArc(dpy,window,xgc,node[i].x-dia/2,node[i].y-dia/2,dia,dia,
		0,64*360);
	memset(ns,0,sizeof(ns));
	switch (i) {
		case 0: ns[0] = 'S'; break;
		case NNODE-1: ns[0] = 'I'; ns[1] = 'A'; break;
		default: ns[0] = '0' + i;
	}
	draw_string(node[i].x,node[i].y,ns);
	if ((k = node[i].next) < NNODE) draw_arrow(i,k,ArcNext);
	if ((k = node[i].daughter) < NNODE) draw_arrow(i,k,ArcDaughter);
}
static void fill_node(i)
int	i;
{
	unsigned long	pixel;
	int	k = node[i].col;
	extern	unsigned long rgb_to_pixel();
	if (k > 200) pixel = rgb_to_pixel(255,(300-k)*255/100,0);
	else if (k > 100) pixel = rgb_to_pixel((k-100)*255/100,255,0);
	else pixel = rgb_to_pixel(0,255,(100-k)*255/100);
	XSetForeground(dpy,xgc,pixel);
	XFillArc(dpy,window,xgc,node[i].x-dia/2,node[i].y-dia/2,dia,dia,
		0,64*360);
}
static void redraw()
{
	int	i;
	extern	Visual	visual;
	if (visual.class == TrueColor) for (i = 0; i < NNODE; i ++)
		if (node[i].active) fill_node(i);
	XSetForeground(dpy,xgc,hidden_pixel);
	for (i = 0; i < NNODE; i ++) if (!node[i].active) draw_node(i);
	XSetForeground(dpy,xgc,fg);
	for (i = 0; i < NNODE; i ++) if (node[i].active) draw_node(i);
}
static void mark_active(i)
int	i;
{
	if (i > 8) return;
	if (node[i].active) return;
	node[i].active = 1;
	mark_active(node[i].next);
	mark_active(node[i].daughter);
}
static void count_states(c) Cell *c;
{
	int k = c->state & 0x0f;
	node[(k >= NNODE)? NNODE-1 : k].col ++;
}
static void init_node()
{
	int	i, k, r = dia * NNODE/4;
	double	th;
	Gene	*g = MonitoredInd->gene;
	char	*m, msg[128];
	sprintf(msg,"#%d: %d cells\n",MonitoredInd->id,MonitoredInd->size);
	for (m = msg + strlen(msg), i = 0; i < GeneLen; i ++, m += 9)
		sprintf(m,"%02x%02x%02x%02x%c",g[i].state,g[i].alpha,
			g[i].beta,g[i].gamma,((i%4==3)?'\n':':'));
	XtVaSetValues(message,XmNvalue,msg,NULL);
	for (i = 0; i < NNODE; i ++) {
		node[i].th = th = i * 2 * M_PI/NNODE;
		node[i].x = (int)(cos(th) * r) + WinWidth/2;
		node[i].y = (int)(sin(th) * r) + WinHeight/2;
		node[i].active = node[i].col = 0;
	}
	for (i = 0; i < 8; i ++) {
		k = g[i].state & 0x0f;
		node[i].next = (k < 8)? k : 8;
		k = g[i].state >> 4;
		node[i].daughter = (k < 8)? k : 8;
	}
	node[8].next = node[8].daughter = NNODE;
	mark_active(0);
	for_all_cells(MonitoredInd,count_states);
	for (i = k = 0; i < NNODE; i ++) if (k < node[i].col) k = node[i].col;
	for (i = 0; i < NNODE; i ++) node[i].col = 300 * node[i].col / k;
}
static void CloseCB(w,cl,cd) Widget w; XtPointer cl,cd; {
	viewg_off();
	XtVaSetValues(GenomeGraphButton,XmNset,False,NULL);
}
static void RedrawCB(w,cl,cd) Widget w; XtPointer cl,cd; {
	XmDrawingAreaCallbackStruct *d = (XmDrawingAreaCallbackStruct *)cd;
	if (d->event->xexpose.count != 0) return;
	redraw();
}
static void DestroyCB(w,cl,cd) Widget w; XtPointer cl,cd; {
	XtVaSetValues(GenomeGraphButton,XmNset,False,NULL);
	shell = NULL;
}	
static Widget create_viewg_shell()
{
	Widget	main, bar, button, draw;
	Arg	args[5];
	int	n;
	XGCValues	xgcv;
	XColor	scol, ecol;
	extern	Colormap	cmap;
	shell = XtCreatePopupShell("ViewGene",
		topLevelShellWidgetClass,app_sh,ShellArgs,NShellArg);
	if (!shell) { perror("cannot create shell widget."); return NULL; }
	n = 0; XtSetArg(args[n],XmNspacing,0); n ++;
	XtSetArg(args[n],XmNmarginHeight,0); n ++;
	XtSetArg(args[n],XmNmarginWidth,0); n ++;
	XtManageChild(main = XmCreateRowColumn(shell,"main",args,n));
	XtManageChild(bar=XmCreateMenuBar(main,"menu_bar",NULL,0));
	XtManageChild(button=XmCreateCascadeButton(bar,"Close",NULL,0));
	XtAddCallback(button,XmNactivateCallback,CloseCB,NULL);
	XtManageChild(draw=XmCreateDrawingArea(main,"ViewGDraw",NULL,0));
	n = 0;
	XtSetArg(args[n],XmNeditable,False); n ++;
	XtSetArg(args[n],XmNeditMode,XmMULTI_LINE_EDIT); n ++;
	XtSetArg(args[n],XmNrows,3); n ++;
	XtManageChild(message=XmCreateText(main,"ViewGMessage",args,n));
	XtVaSetValues(main,XmNmessageWindow,message,NULL);
	XtAddCallback(shell,XmNdestroyCallback,DestroyCB,NULL);
	XtPopup(shell,XtGrabNone);
	XtVaGetValues(draw,XmNwidth,&WinWidth,XmNheight,&WinHeight,
		XmNforeground,&fg,XmNbackground,&bg,NULL);
	if (!(window = XtWindow(draw)))
		{ perror("cannot find window."); return NULL; }
	if (!(fst = XLoadQueryFont(dpy, STRINGFONT))) {
		fprintf(stderr,"%s not found.\n",STRINGFONT);
		fst = XLoadQueryFont(dpy,"variable");
		if (!fst) exit(1);
	}
	BaseOffset = (fst->max_bounds.ascent+fst->max_bounds.descent)
		/2 - fst->max_bounds.descent;
	hidden_pixel = XAllocNamedColor(dpy,cmap,HIDDENCOLOR,&scol,&ecol)?
		scol.pixel : bg;
	xgcv.foreground = fg;
	xgcv.background = bg;
	xgcv.line_width = 2;
	xgcv.line_style = LineSolid;
	xgcv.font = fst->fid;
	xgc = XCreateGC(dpy,window,GCForeground|GCBackground|GCLineWidth|
		GCLineStyle|GCFont,&xgcv);
	dia = (int)((WinWidth>WinHeight)?WinHeight:WinWidth)*6/5/NNODE;
	init_node();
	XtAddCallback(draw,XmNexposeCallback,RedrawCB,NULL);
	return shell;
}
viewg_on(in)
Individual	*in;
{
	MonitoredInd = in? in : Population;
	if (!shell) { shell = create_viewg_shell(); if (!shell) return 0; }
	else {
		init_node();
		XtPopup(shell,XtGrabNone);
		XSetForeground(dpy,xgc,bg);
		XFillRectangle(dpy,window,xgc,0,0,WinWidth,WinHeight);
		redraw();
	}
	return 1;
}
viewg_off()
{
	MonitoredInd = NULL; XtPopdown(shell);
}
viewg(in)
Individual	*in;
{
	extern	Widget	GenomeGraphButton;
	if (viewg_on(in)) XtVaSetValues(GenomeGraphButton,XmNset,True,NULL);
}
